--*******************************************************************************************************************************
--
-- Description: This module acquires data and over-range status from a variety of A/D converter(ADC) chips.  The bus type
--				supports multiple RX channels and stuffing of multiple samples per clock cyle to accomodate DDR inputs.  When
--				samples are stuffed, they are chronologically ordered from LSB to MSB.
--					
-- Rev History: 11-28-2018:	Added MAX_TAP_VALUE generic to the modified clock lock entity.
-- 				01-10-2018:	Replaced asynchronous reset with synchronous reset.
--				08-04-2016:	Added rx_frame_in port an eliminated unused clock input and changed the rx_clk_in data type.
--				02-18-2016:	Modified code to support multiplexing two channels on one input bus and changed over-range port type.
--				07-14-2015:	Replaced IDELAY components with IDELAY2E which also required an update to clock_lock.
-- 				07-30-2014: Added logic to accommodate clock lock detecting either edge of a DDR clock.
--				09-04-2013: Initial Release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all; 

Library UNISIM; 
use UNISIM.Vcomponents.all;

use work.model_specific_package.all;
use work.transceiver_package.all;

entity adc_interface is 
	port (
	--  Reset and clocks supplied by the application logic.
    rx_srst			: in std_logic;		-- Active high synchronous reset in the RX clock domain.
	rx_clk			: in std_logic;		-- Clock used to register ADC sample data.
	
	-- ADC data supplied to the application logic.
	rx_data_out		: out RX_BUS;		-- Data bus carrying ADC samples.
	rx_ovr_out		: out OVR_BIT;		-- Logic one indicates an over-range condition on the ADC.

	-- Serial clock lock command signal supplied from/to the application logic.
	clk_lck_start	: in std_logic; 	-- Initiates the clock lock sequence that sets IODELAY values for data inputs.
	clk_lck_busy	: out std_logic; 	-- Logic one indicates that the clock lock process is active. 
	clk_lck_status	: out std_logic_vector(ADC_CLK_SOURCES-1 downto 0); -- Logic one indicates that clocks are locked.
	
	--	The following signals connect to ADC devices through FPGA pins.
	rx_data_in		: in RX_BUS_IN;		-- ADC data bus.
	rx_frame_in  	: in RX_FRAME_BIT;	-- ADC channel frame sync.
	rx_ovr_in		: in OVR_BIT_IN;	-- ADC over-range flag.
	rx_clk_in		: in RX_CLOCK_BIT	-- ADC output clock.
);	
end adc_interface;

architecture behavioral of adc_interface is

	signal rx_data_dly		: RX_BUS_IN;	-- Time aligned data bus from the IDELAY primitive.
	signal rx_data_re		: RX_BUS_IN;	-- Data bus carrying ADC samples (DDR rising edge).
	signal rx_data_fe		: RX_BUS_IN;	-- Data bus carrying ADC samples (DDR falling edge).
	signal rx_data_fe_old	: RX_BUS_IN;	-- Falling edge data samples stored for possible DDR realignment.
	signal rx_ovr_dly		: OVR_BIT_IN;	-- Time aligned over-range bit from the IDELAY primitive.
	signal rx_ovr_re		: OVR_BIT_IN;	-- Logic one indicates an over-range condition on the ADC (DDR rising edge).
	signal rx_ovr_fe		: OVR_BIT_IN;	-- Logic one indicates an over-range condition on the ADC (DDR falling edge).
	signal rx_ovr_fe_old	: OVR_BIT_IN;	-- Falling edge over-range bit stored for possible DDR realignment.
	signal rx_clk_dly		: RX_1;			-- Time aligned data ready bit from the IDELAY primitive.
	signal adc_stuffed		: RX_BUS;		-- Potentially multiple ADC samples stuffed on a single bus.
	signal ovr_stuffed		: OVR_BIT;		-- Potentially multiple over range stuffed on a single bus.
	signal rx_clk_re		: std_logic_vector(ADC_CLK_SOURCES-1 downto 0);	-- Registered ADC forwarded clock signal (DDR rising edge).
	signal rx_clk_fe		: std_logic_vector(ADC_CLK_SOURCES-1 downto 0);	-- Registered ADC forwarded clock signal (DDR falling edge).
	signal clk_edge_detect	: std_logic_vector(ADC_CLK_SOURCES-1 downto 0);	-- Reports which edge was detected, rising (0) or falling (1).
	signal fwd_counter		: std_logic_vector((5*ADC_CLK_SOURCES)-1 downto 0);		-- Tap count reported by the IDELAY attached to the forwarded clock.
	signal fwd_ld			: std_logic_vector(ADC_CLK_SOURCES-1 downto 0);		-- Logic one loads the starting tap value in the IDELAY attached to the forwarded clock.
	signal fwd_ce			: std_logic_vector(ADC_CLK_SOURCES-1 downto 0);		-- Logic one enables the increment/decrement function in the IDELAY attached to the forwarded clock.
	signal fwd_inc			: std_logic_vector(ADC_CLK_SOURCES-1 downto 0);		-- Logic one increments the tap value and logic zero decrements the tap value in the IDELAY attached to the forwarded clock.
	signal data_ld			: std_logic_vector(ADC_CLK_SOURCES-1 downto 0);		-- Logic one loads the starting tap value in the IDELAY attached to the data bits.
	signal data_ce			: std_logic_vector(ADC_CLK_SOURCES-1 downto 0);		-- Logic one enables the increment/decrement function in the IDELAY attached to the data bits.
	signal data_inc			: std_logic_vector(ADC_CLK_SOURCES-1 downto 0);		-- Logic one increments the tap value and logic zero decrements the tap value in the IDELAY attached to the data bits.
	
begin
 	
--************************************************ INTERNAL SIGNAL ASSIGNMENTS *************************************************

	-- Sequence through each channel to process data bits.
	ADC_CH : for i in 1 to zero_to_one(RX_CHANNEL_COUNT/RX_DATA_SHARE) generate
		-- Register the falling edge value for possible DDR timing realignment.
		process (rx_clk)
		begin
			if rising_edge(rx_clk) then
				rx_data_fe_old(i) <= rx_data_fe(i);
			end if;
		end process;
		-- Sequence through each data item bit that gets created in a single clock cycle.
		ADC_BIT : for j in 0 to (RX_ITEM_PER_CLK*RX_ITEM_SIZE/RX_INTERLEAVE)-1 generate
			-- There is always new data on the rising clock edge whether data is stuffed (DDR inputs) or not (SDR inputs).
			adc_stuffed((i*RX_DATA_SHARE)-integer(RX_DATA_SHARE/2))(j*RX_INTERLEAVE) <= rx_data_fe_old(i)(j) when clk_edge_detect(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))) = '1' else rx_data_re(i)(j);
			-- If bits are interleaved (DDR), there is also new data on the falling clock edge.
			BIT_INTERLEAVE: if RX_INTERLEAVE > 1 generate
				adc_stuffed(i)((j*RX_INTERLEAVE)+1) <= rx_data_re(i)(j) when clk_edge_detect(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))) = '1' else rx_data_fe(i)(j);
			end generate;
			-- If items are stuffed (DDR), there is also new data on the falling clock edge.
			ITEM_STUFF: if RX_ITEM_STUFF > 1 generate
				adc_stuffed(i)(j+(RX_ITEM_PER_CLK*RX_ITEM_SIZE)) <= rx_data_re(i)(j) when clk_edge_detect(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))) = '1' else rx_data_fe(i)(j);
			end generate;
			-- If channels share an input port, there is also new data on the falling clock edge.
			PORT_SHARE: if RX_DATA_SHARE > 1 generate
				adc_stuffed(i*RX_DATA_SHARE)(j) <= rx_data_re(i)(j) when clk_edge_detect(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))) = '1' else rx_data_fe(i)(j);
			end generate;
		end generate;
	end generate;
	
	-- Sequence through each channel to process over-range bits.
	OVR_CH : for i in 1 to zero_to_one(RX_CHANNEL_COUNT/RX_DATA_SHARE) generate
		-- Register the falling edge value for possible DDR timing realignment.
		process (rx_clk)
		begin
			if rising_edge(rx_clk) then
				rx_ovr_fe_old(i) <= rx_ovr_fe(i);
			end if;
		end process;
		-- Sequence through each data item that gets created in a single clock cycle.
		OVR_BIT : for j in 0 to RX_ITEM_PER_CLK-1 generate
			-- There is always a new over-range flag on the rising clock edge whether data is stuffed (DDR inputs) or not (SDR inputs).
			-- Note: Replicate a single over-range bit across multiple stuffed samples if it is shared.
			ovr_stuffed((i*RX_DATA_SHARE)-integer(RX_DATA_SHARE/2))(j) <= rx_ovr_fe_old(i)(j/RX_ITEM_PER_OVR) when clk_edge_detect(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))) = '1' else rx_ovr_re(i)(j/RX_ITEM_PER_OVR);
			-- If items are stuffed (DDR), there is also a new over-range flag on the falling clock edge.
			OVR_STUFF: if RX_ITEM_STUFF > 1 generate
				-- Note: Replicate a single over-range bit across multiple stuffed samples if it is shared.
				ovr_stuffed(i)(j+RX_ITEM_PER_CLK) <= rx_ovr_re(i)(j/RX_ITEM_PER_OVR) when clk_edge_detect(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))) = '1' else rx_ovr_fe(i)(j/RX_ITEM_PER_OVR);
			end generate;
			-- If channels share an input port, there is also a new over-range bit on the falling clock edge.
			OVR_SHARE: if RX_DATA_SHARE > 1 generate
				ovr_stuffed(i*RX_DATA_SHARE)(j) <= rx_ovr_re(i)(j) when clk_edge_detect(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))) = '1' else rx_ovr_fe(i)(j);
			end generate;
		end generate;
	end generate;

	-- Register the data and over-range bits to improve timing performance.
	process (rx_clk)
	begin
		if rising_edge(rx_clk) then
			rx_data_out <= adc_stuffed;
			rx_ovr_out <= ovr_stuffed;
		end if;
	end process;
	
--*************************************************** CLOCK LOCK CONTROLLER ****************************************************

    -- Optimize timing at the SRAM read port.
    ADC_CLK_LCK : entity work.clock_lock
    generic map (
		SOURCES			=> ADC_CLK_SOURCES,
		CLK_FWD_INIT	=> MAX_TAP_VALUE,
		MAX_SEARCH		=> ADC_MAX_SEARCH,
		RESET_IF_NOSOL	=> ADC_RESET_IF_NOSOL
	)
	port map (
		srst         	=> rx_srst,
		lock_start   	=> clk_lck_start,
		lock_busy   	=> clk_lck_busy,
		lock_status    	=> clk_lck_status,
		clk_lck_clk		=> rx_clk,
		clk_fwd			=> rx_clk_re,
		clk_edge_search	=> ADC_CLK_EDGE,
		clk_edge_detect	=> clk_edge_detect,
		fwd_counter		=> fwd_counter,
		fwd_ld			=> fwd_ld,
		fwd_ce			=> fwd_ce,
		fwd_inc			=> fwd_inc,
		data_ld			=> data_ld,
		data_ce			=> data_ce,
		data_inc		=> data_inc
	);
	
--************************************************** EXTERNAL PIN ASSIGNMENTS **************************************************

	-- Instantiate the requested number of receiver channels.
	ADC_CHANNEL : for i in 1 to zero_to_one(RX_CHANNEL_COUNT/RX_DATA_SHARE) generate
		-- Process each ADC data item bit.	
		ADC_DATA : for j in 0 to (RX_ITEM_PER_CLK*RX_ITEM_SIZE/RX_INTERLEAVE)-1 generate
			-- Set the IDELAY values for each data bit using the clock lock controller.
			DATA_IDELAY : IDELAYE2
				generic map(IDELAY_TYPE => "VARIABLE", IDELAY_VALUE => ADC_DATA_INIT(((i-1)*RX_ITEM_SIZE/RX_INTERLEAVE)+j), HIGH_PERFORMANCE_MODE => "TRUE")
				port map(C => rx_clk, REGRST => '0', LD => data_ld(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))), CE => data_ce(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))), INC => data_inc(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))), CINVCTRL => '0', CNTVALUEIN => "00000", IDATAIN => rx_data_in(i)(j), DATAIN => '0', LDPIPEEN => '0', DATAOUT => rx_data_dly(i)(j), CNTVALUEOUT => open);
			-- Use double data rate (DDR) input registers in the IOB to capture data on the rising and falling edge of the sample clock.
			-- Note: The DDR register is used even on SDR inputs to ensure that the signal is registered inside the IOB.
			DATA_DDR : IDDR
				generic map(DDR_CLK_EDGE => "SAME_EDGE_PIPELINED") 
				port map(Q1 => rx_data_re(i)(j), Q2 =>rx_data_fe(i)(j), C => rx_clk, CE => '1', D => rx_data_dly(i)(j), R => '0', S => '0');
		end generate;
	end generate;

	-- Map the available over-range indicators to each channel.
	ADC_OVR : for i in 1 to zero_to_one(RX_CHANNEL_COUNT/RX_DATA_SHARE) generate
		ADC_OVR : for j in 0 to (RX_ITEM_PER_CLK/RX_ITEM_PER_OVR)-1 generate
			--  Set the IDELAY value for the over-range signals using the clock lock controller. 
			OVR_IDELAY : IDELAYE2
				generic map(IDELAY_TYPE => "VARIABLE", IDELAY_VALUE => ADC_OVR_INIT, HIGH_PERFORMANCE_MODE => "TRUE")
				port map(C => rx_clk, REGRST => '0', LD => data_ld(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))), CE => data_ce(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))), INC => data_inc(integer((i-1)*ADC_CLK_SOURCES/zero_to_one(RX_CHANNEL_COUNT))), CINVCTRL => '0', CNTVALUEIN => "00000", IDATAIN => rx_ovr_in(i)(j), DATAIN => '0', LDPIPEEN => '0', DATAOUT => rx_ovr_dly(i)(j), CNTVALUEOUT => open);
			-- Use double data rate (DDR) input registers in the IOB to capture data on the rising and falling edge of the sample clock.
			-- Note: The DDR register is used on SDR inputs to ensure that the signal is registered inside the IOB.
			OVR_DDR : IDDR
				generic map(DDR_CLK_EDGE => "SAME_EDGE_PIPELINED")
				port map(Q1 => rx_ovr_re(i)(j), Q2 => rx_ovr_fe(i)(j), C => rx_clk, CE => '1', D => rx_ovr_dly(i)(j), R => '0', S => '0');
		end generate;
	end generate;
	
	-- Process each clock forwarded by the ADC.	
	CLK_FORWARD : for i in 1 to ADC_CLK_SOURCES generate
		--  Set the IDELAY value for the clock forward signals using the clock lock controller.
		CLK_IDELAY : IDELAYE2
			generic map(IDELAY_TYPE => "VARIABLE", IDELAY_VALUE => ADC_CLK_INIT, HIGH_PERFORMANCE_MODE => "TRUE")
			port map(C => rx_clk, REGRST => '0', LD => fwd_ld(i-1), CE => fwd_ce(i-1), INC => fwd_inc(i-1), CINVCTRL => '0', CNTVALUEIN => "00000", IDATAIN => rx_clk_in(i), DATAIN => '0', LDPIPEEN => '0', DATAOUT => rx_clk_dly(i), CNTVALUEOUT => fwd_counter((5*i)-1 downto 5*(i-1)));
		-- Use double data rate (DDR) input registers in the IOB to capture data on the rising and falling edge of the sample clock.
		-- Note: The DDR register is used on SDR inputs to ensure that the signal is registered inside the IOB.
		CLK_DDR : IDDR
			generic map(DDR_CLK_EDGE => "SAME_EDGE_PIPELINED")
			port map(Q1 => rx_clk_re(i-1), Q2 => open, C => rx_clk, CE => '1', D => rx_clk_dly(i), R => '0', S => '0');
	end generate;
		
end behavioral;
