--*******************************************************************************************************************************
--
-- Description: This module measures an active clock frequency using a software timer.
--
-- Rev History: 01-10-2018:	Initial release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

Library UNISIM;
use UNISIM.vcomponents.all;
library UNIMACRO;
use unimacro.Vcomponents.all;

entity clock_frequency is 
	port (
	-- Clocks.
    clk                 : in std_logic;						-- System clock.
	clk_to_measure		: in std_logic;						-- IOB sample clock.
    
    -- Clock frequency measurement command and result.
    clk_measurement		: in std_logic_vector(1 downto 0);	-- Logic one commands the start (0) and stop (1) of a frequency measurement.
    clk_frequency		: out std_logic_vector(29 downto 0)	-- Measured TX clock frequency.
	);	
	end clock_frequency;

architecture behavioral of clock_frequency is

	signal freq_count		: std_logic_vector(1 downto 0);	-- Active high enable to the frequency counter.
	signal freq_on_off		: std_logic_vector(1 downto 0);	-- Active high on/off (bit 0/bit 1) switch for the frequency measurement.

begin                    

	-- Create a single clock period on/off (bit 0/bit 1) command to the frequency counter in the TX clock domain.
	FREQ_CMD : for i in 0 to 1 generate
		ON_OFF : entity work.sync_reset
		port map (
			source_clk		=> clk,
			source_rst		=> clk_measurement(i),
			target_clk  	=> clk_to_measure,
			target_rst		=> freq_on_off(i)
		);
	end generate;
	
	-- Set a sticky bit to enable the frequency counter over the interval commanded by software.
	process (clk_to_measure)
	begin
		if rising_edge(clk_to_measure) then
			if freq_on_off(0) = '1' then
				freq_count(0) <= '1';
			elsif freq_on_off(1) = '1' then
				freq_count(0) <= '0';
			end if;
	   end if;
	end process;
	
	-- Enable the frequency measurement clock when the counter is reset or commanded to make a measurement.
	freq_count(1) <= freq_on_off(0) or freq_count(0);
		
	-- Count the number of TX clock periods over the interval commanded by software.
	-- Note: Clock frequency measurement is needed to compute PLL settings in the DAC.
	FREQ_COUNTER : COUNTER_LOAD_MACRO
		generic map (DEVICE => "7SERIES", COUNT_BY => x"000000000001", WIDTH_DATA => 30)
		port map (Q => clk_frequency, CLK => clk_to_measure, CE => freq_count(1), DIRECTION => '1', LOAD => '0', LOAD_DATA => (others => '0'), RST => freq_on_off(0));
	
end behavioral;
