--*******************************************************************************************************************************
--
-- Description: This module sends data to a variety of D/A converter(DAC) chips.  The bus type supports multiple TX channels
--				and stuffing of multiple samples per clock cyle to accomodate DDR outputs.  When samples are stuffed, they are
--				chronologically ordered from LSB to MSB.
--					
-- Rev History: 01-10-2018:	Replaced asynchronous reset with synchronous reset.
--				08-04-2016 Added support for frame sync signal.
-- 				09-04-2013 Initial Release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all; 

Library UNISIM; 
use UNISIM.Vcomponents.all;

use work.model_specific_package.all;
use work.transceiver_package.all;

entity dac_interface is 
	port (
    -- Reset and clock.                                                                                      
	tx_clk			: in std_logic;		-- Clock used to register DAC sample data.
	
	-- DAC data supplied from the application logic.
	tx_data_in		: in TX_BUS;		-- Data bus carrying DAC samples.
	tx_frame_in		: in TX_1;	-- Frame sync.
	
	-- The following signals connect to DAC devices through FPGA pins.
	tx_data_out		: out TX_BUS_OUT;	-- Data bus carrying DAC samples.
	tx_frame_out	: out TX_FRAME_BIT;	-- Frame sync.
	tx_clk_out		: out TX_CLOCK_BIT	-- Internal single ended version of DAC forwarded clock.

	);	
	end dac_interface;	

architecture behavioral of dac_interface is

begin                    
 	
--************************************************** EXTERNAL PIN ASSIGNMENTS **************************************************

	-- Instantiate the requested number of transmitter channels.
	DAC_CHANNEL : for i in RX_CHANNEL_COUNT+1 to integer(TX_CHANNEL_COUNT/TX_DATA_SHARE)+RX_CHANNEL_COUNT generate
	
		DAC_DATA : for j in 0 to TX_ITEM_SIZE-1 generate
			-- Use double data rate (DDR) output registers in the IOB to transmit data on the rising and falling edge of the sample clock.
			-- Note: The DDR register is used on SDR outputs to ensure that the signal is registered inside the IOB.
			DATA_DDR : ODDR
				GENERIC MAP(DDR_CLK_EDGE => "SAME_EDGE")
				PORT MAP(D1 => tx_data_in((TX_DATA_SHARE*i)-integer(TX_DATA_SHARE/2)-RX_CHANNEL_COUNT)(j), D2 => tx_data_in((TX_DATA_SHARE*i)-RX_CHANNEL_COUNT)(j), C => tx_clk, CE => '1', Q => tx_data_out(i)(j), R => '0', S => '0');
		end generate;
	
		-- Instantiate the frame sync.
		DAC_FRAME : if (i-RX_CHANNEL_COUNT-1) mod TX_FRAME_SHARE = 0 generate
			-- Use a double data rate (DDR) register to generate the frame sync.
			-- Note: The DDR register is used on SDR outputs to ensure that the signal is registered inside the IOB.
			FRAME_DDR : ODDR
				generic map (DDR_CLK_EDGE => "SAME_EDGE")
				port map (D1 => tx_frame_in((TX_DATA_SHARE*i)-integer(TX_DATA_SHARE/2)-RX_CHANNEL_COUNT), D2 => tx_frame_in((TX_DATA_SHARE*i)-RX_CHANNEL_COUNT), C => tx_clk, CE => '1', Q => tx_frame_out(i), R => '0', S => '0');
		end generate;

	end generate;
	
	-- Instantiate all of the required DAC clocks.
	CLK_FORWARD : for i in RX_CHANNEL_COUNT+1 to DAC_CLK_SOURCES+RX_CHANNEL_COUNT generate
		-- Use a double data rate (DDR) register to generate the forwarded DAC clock.
		CLK_DDR : ODDR
		generic map (DDR_CLK_EDGE => "SAME_EDGE") 
		port map (D1 => '1', D2 => '0', C => tx_clk, CE => '1', Q => tx_clk_out(i), R => '0', S => '0');
	end generate;
			
end behavioral;
