--*******************************************************************************************************************************
--
-- Description: This module maps the requested number of DMA channels across the defined number of DMA engines.  Channels will
--				share DMA engines if necessary.	There may be DMA channels used for purposes other than transfer of raw ADC/DAC data.
--				Note: It is assumed that DMA_ENGINE_COUNT <= DMA_CHANNEL_COUNT.
--					
-- Rev History: 01-10-2018:	Initial release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all; 
use IEEE.math_real.all; 

library UNISIM; 
use UNISIM.Vcomponents.all; 

use work.model_specific_package.all;
use work.dma_package.all;

entity pci_extended_dma is
	port (
    -- Reset and clock.                                                                                      
    srst_ext			: in std_logic;			-- Active high synchronous reset in the system clock domain.
	clk_ext 			: in std_logic;			-- System clock supplied by the endpoint core.

	-- DMA control and status signals.
	dma_wrchannel 	: out std_logic_vector(DMA_CHANNEL_COUNT downto 0);	-- Refer to the PLDA EZDMA2 reference manual.
	dma_rdchannel 	: out std_logic_vector(DMA_CHANNEL_COUNT downto 0);	-- Refer to the PLDA EZDMA2 reference manual.
	dma_regin		: in DMA_TYPE_128;						-- Refer to the PLDA EZDMA2 reference manual.
	dma_param		: in DMA_TYPE_24;						-- Refer to the PLDA EZDMA2 reference manual.
	dma_control		: in DMA_TYPE_6;						-- Refer to the PLDA EZDMA2 reference manual.
	dma_status		: out DMA_TYPE_4;						-- Refer to the PLDA EZDMA2 reference manual.
	dma_fifocnt		: in DMA_TYPE_13;						-- Refer to the PLDA EZDMA2 reference manual.
	dma_rddata		: in DMA_TYPE_AXI;						-- Refer to the PLDA EZDMA2 reference manual.
	dma_ext_active	: out DMA_TYPE_1;						-- Logic one indicates that the extended DMA controller is processing a request.

	-- Extended DMA control and status signals.
	dma_wrchannel_ext 	: in std_logic_vector(DMA_ENGINE_COUNT downto 0);	-- Extended version of dma_regin (above).
	dma_rdchannel_ext 	: in std_logic_vector(DMA_ENGINE_COUNT downto 0);	-- Extended version of dma_regin (above).
	dma_regin_ext		: out HECTA_128;					-- Extended version of dma_regin (above).
	dma_regout_ext		: in HECTA_128;						-- Refer to the PLDA EZDMA2 reference manual.
	dma_param_ext		: out HECTA_24;						-- Extended version of dma_param (above).
	dma_control_ext		: out HECTA_6;						-- Extended version of dma_control (above).
	dma_status_ext		: in HECTA_4;						-- Extended version of dma_status (above).
	dma_fifocnt_ext		: out HECTA_13;						-- Extended version of dma_fifocnt (above).
	dma_rddata_ext		: out HECTA_AXI						-- Extended version of dma_rddata (above).
	);
end pci_extended_dma;

architecture structural of pci_extended_dma is

	type arbstate is (cycle, check, latency, request, connect, monitor, store);	-- Type definitions for the arbiter state machine.
	type arb_array is array (DMA_ENGINE_COUNT downto 1) of arbstate;	-- Array of arbiter state machines.
	signal arb_state		: arb_array;  		-- Current arbiter machine state.
	signal cycle_count		: DMA_EXT_SHAREU;	-- Counter used to sequentially cycle through all DMA channels sharing an engine.
	signal desc_address		: DMA_EXT_TYPE_8;	-- Address to descriptor memory.
	signal desc_data		: DMA_EXT_TYPE_64;	-- Output data from descriptor memory.
	signal desc_write		: DMA_EXT_TYPE_0;	-- Active high write enable to descriptor memory.
	signal dma_pending		: DMA_EXT_PEND_6;	-- Control inputs to the arbiters, including residual channels.
	signal dma_request		: DMA_EXT_TYPE_6;
	
begin

--********************************************* DEDICATED AND UNUSED DMA ENGINES ***********************************************

	-- Make direct connections to channels that do not share a DMA engine.
	dma_wrchannel(0) <= dma_wrchannel_ext(0);
	dma_rdchannel(0) <= dma_rdchannel_ext(0);
	-- Direct connections are possible if the number of channels is less than twice the available DMA engines.
	DMA_DIRECT : if DMA_CHANNEL_COUNT < 2*DMA_ENGINE_COUNT generate
		DMA_ENGINE : for i in 1 to DMA_ENGINE_COUNT generate
			DIRECT_CONNECT : if i > DMA_CHANNEL_COUNT-DMA_ENGINE_COUNT generate
				dma_regin_ext(i) <= dma_regin(i);
				dma_param_ext(i) <= dma_param(i);
				dma_control_ext(i) <= dma_control(i);
				dma_fifocnt_ext(i) <= dma_fifocnt(i);
				dma_rddata_ext(i) <= dma_rddata(i);
				dma_wrchannel(i) <= dma_wrchannel_ext(i);
				dma_rdchannel(i) <= dma_rdchannel_ext(i);
				dma_status(i) <= dma_status_ext(i);
				dma_ext_active(i) <= '1';
			end generate;
		end generate;
	end generate;

	-- Make signal assignments to unused DMA engines.
	DMA_UNUSED : for i in DMA_ENGINE_COUNT+1 to 7 generate
		dma_regin_ext(i) <= (others=>'0');
		dma_param_ext(i) <= (others=>'0');
		dma_control_ext(i) <= (others=>'0');
		dma_fifocnt_ext(i) <= (others=>'0');
		dma_rddata_ext(i) <= (others=>'0');
	end generate;
	
	-- Map the DMA control inputs from the chanels to the arbiters and create terminated inputs for residual arbiter inputs.
	DMA_REQ : for i in 1 to DMA_ENGINE_COUNT*natural(ceil(real(DMA_CHANNEL_COUNT)/real(DMA_ENGINE_COUNT))) generate
		REQ_DIRECT : if i <= DMA_CHANNEL_COUNT generate
			dma_pending(i) <= dma_control(i);
		end generate;
		REQ_UNUSED : if i > DMA_CHANNEL_COUNT generate
			dma_pending(i) <= (others=>'0');
		end generate;
	end generate;

--************************************************* STATE MACHINE TRANSITIONS **************************************************

	-- Create an arbiter state machine for shared DMA engines.
	DMA_ARB_STATE : for i in 1 to DMA_ENGINE_COUNT generate
		-- Determine whether each DMA engine is going to service more than one DMA channel.
		DMA_SHARE_STATE : if i <= DMA_CHANNEL_COUNT-DMA_ENGINE_COUNT generate
	
			process(clk_ext,srst_ext)
			begin
				if rising_edge(clk_ext) then
				
					if srst_ext = '1' then
						arb_state(i) <= cycle;
						
					else
						-- Define each of the arbiter state machine transitions.
						case arb_state(i) is
						
							-- Cycle through all of the connected DMA channels in sequence.
							when cycle =>
								arb_state(i) <= check;
								
							-- Check if this DMA channel is requesting a transaction.
							when check =>
								-- No DMA transaction request pending.
								if dma_request(i) = "000000" then
									arb_state(i) <= cycle;
								-- Connect the requesting channel to the DMA engine.
								else
									arb_state(i) <= latency;
								end if;
								
							-- Wait an extra clock to accommodate the two cycle read latency of the descriptor memory.
							when latency =>
								arb_state(i) <= request;
								
							-- Pass the request to the assigned DMA engine.
							when request =>
								arb_state(i) <= connect;
								
							-- Wait for an acknowledgement to the DMA request.
							when connect =>
								if dma_status_ext(i)(3) = '1' then
									arb_state(i) <= monitor;
								end if;
								
							-- Wait for the DMA burst transfer to complete.
							when monitor =>
								-- Wait for the DMA status bit to transition from busy to idle.
								if dma_status_ext(i)(3) = '0' then
									arb_state(i) <= store;
								end if;
							
							-- Store the current descriptor data.
							when store =>
								arb_state(i) <= cycle;
								
							-- Define the default state to be cycle.
							when others => arb_state(i) <= cycle;
							
						end case;
					end if;
				end if;
			end process;

		end generate;
	end generate;

--*************************************************** STATE MACHINE EVENTS *****************************************************

	DMA_ENG : for i in 1 to DMA_ENGINE_COUNT generate
		-- Determine whether each DMA engine is going to service more than one DMA channel.
		SHARE_ENG : if i <= DMA_CHANNEL_COUNT-DMA_ENGINE_COUNT generate
	
			process(clk_ext,srst_ext)
			begin
				if rising_edge(clk_ext) then
				
					if srst_ext = '1' then
						cycle_count(i) <= (others => '0');
						
					else
						-- Increment the cycle counter to sequence through all of the channels connected to a DMA engine.
						if arb_state(i) = cycle then
							if cycle_count(i) = to_unsigned(natural(ceil(real(DMA_CHANNEL_COUNT)/real(DMA_ENGINE_COUNT)))-1,cycle_count(i)'length) then
								cycle_count(i) <= (others=>'0');
							else
								cycle_count(i) <= cycle_count(i) + to_unsigned(1,cycle_count(i)'length);
							end if;
						end if;
						
					end if;
				end if;
			end process;

			-- Store DMA addresses on shared channels since the built-in DMA engine address counters cannot be shared between buffers.
			desc_address(i) <= std_logic_vector(to_unsigned(0,8-cycle_count(i)'length)) & std_logic_vector(cycle_count(i));
			desc_write(i) <= "1" when arb_state(i) = store else "0";
			DESC_EXT : entity work.dma_ext_memory
				port map(
					clka => clk_ext,
					wea => desc_write(i),
					addra => desc_address(i),
					dina => dma_regout_ext(i)(63 downto 0),
					douta => desc_data(i)
					);
	
			-- Connect the assigned DMA engine to the active channel while a transaction is processed.
			dma_request(i) <= dma_pending(i+(DMA_ENGINE_COUNT*to_integer(cycle_count(i))));
			dma_param_ext(i) <= dma_param(i+(DMA_ENGINE_COUNT*to_integer(cycle_count(i))));
			dma_fifocnt_ext(i) <= dma_fifocnt(i+(DMA_ENGINE_COUNT*to_integer(cycle_count(i))));
			dma_rddata_ext(i) <= dma_rddata(i+(DMA_ENGINE_COUNT*to_integer(cycle_count(i))));
			-- A new descriptor is loaded on every transaction requested by a shared channel.
			dma_control_ext(i) <= "011111" when arb_state(i) = request else (others=>'0');
			-- The address is only supplied from the DMA on Demand controller at the start of a new page.
			dma_regin_ext(i)(63 downto 0) <= dma_regin(i+(DMA_ENGINE_COUNT*to_integer(cycle_count(i))))(63 downto 0) when dma_request(i) = "011111" else desc_data(i);
			dma_regin_ext(i)(127 downto 64) <= dma_regin(i+(DMA_ENGINE_COUNT*to_integer(cycle_count(i))))(127 downto 64);
			
			DMA_CHAN : for j in 1 to DMA_CHANNEL_COUNT generate
				SHARE_CHAN : if (j-i) mod DMA_ENGINE_COUNT = 0 generate
					-- Route the shared status bits to each DMA engine
					dma_status(j) <= dma_status_ext(i);
					-- Connect the active channel to the assigned DMA engine while a transaction is processed.
					dma_ext_active(j) <= '1' when to_integer(cycle_count(i)) = (j-1)/DMA_ENGINE_COUNT and arb_state(i) = request else '0';
					dma_wrchannel(j) <= dma_wrchannel_ext(i) when to_integer(cycle_count(i)) = (j-1)/DMA_ENGINE_COUNT else '0';
					dma_rdchannel(j) <= dma_rdchannel_ext(i) when to_integer(cycle_count(i)) = (j-1)/DMA_ENGINE_COUNT else '0';
				end generate;
			end generate;

		end generate;
	end generate;
	
end structural;
